// camera class
// -Kyoung Shin Park (2013 Spring)
//////////////////////////////////////////////////////////////////////
#include "camera.h"
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <iostream>

namespace moglclass
{

// OpenGL world frame is equal to camera frame
Camera::Camera(CameraType type)
{
	cameraType = type;

	pos   = glm::vec3(0.0f, 0.0f, 0.0f);
	right = glm::vec3(1.0f, 0.0f, 0.0f);
	up    = glm::vec3(0.0f, 1.0f, 0.0f);
	look  = glm::vec3(0.0f, 0.0f, 1.0f);

	V = View();
}

Camera::~Camera()
{
}

void Camera::setCameraType(CameraType type_)
{
	cameraType = type_;
}

void Camera::setPosition(float x, float y, float z)
{
	pos[0] = x; pos[1] = y; pos[2] = z; 
}

void Camera::setPosition(const glm::vec3& pos_)
{
	pos = pos_;
}

void Camera::setAt(float x, float y, float z)
{
	look[0] = pos[0] - x; 
	look[1] = pos[1] - y; 
	look[2] = pos[2] - z;
	std::cout << "look: " << look[0] << " " << look[1] << " " << look[2] << std::endl;
}

void Camera::setAt(const glm::vec3& at_)
{
	look = pos - at_;
	std::cout << "look: " << look[0] << " " << look[1] << " " << look[2] << std::endl;
}

void Camera::setRight(float x, float y, float z)
{
	right[0] = x; right[1] = y; right[2] = z; 
}

void Camera::setRight(const glm::vec3& right_)
{
	right = right_;
}

void Camera::setUp(float x, float y, float z)
{
	up[0] = x; up[1] = y; up[2] = z; 
}

void Camera::setUp(const glm::vec3& up_)
{
	up = up_;
}

void Camera::setLook(float x, float y, float z)
{
	look[0] = x; look[1] = y; look[2] = z; 
}

void Camera::setLook(const glm::vec3& look_)
{
	look = look_;
}

glm::vec3 Camera::getPosition(void)
{
	return pos;
}

glm::vec3 Camera::getRight(void)
{
	return right;
}

glm::vec3 Camera::getUp(void)
{
	return up;
}

glm::vec3 Camera::getLook(void)
{
	return look;
}

void Camera::walk(float units)
{
	// move around z-axis
	// -z to move forward in OpenGL
	if( cameraType == DRIVE )
		pos -= glm::vec3(look[0], 0.0f, look[2]) * units;

	if( cameraType == FLY )
		pos -= look * units;
}

void Camera::strafe(float units)
{
	// move around x-axis
	if( cameraType == DRIVE )
		pos += glm::vec3(right[0], 0.0f, right[2]) * units;

	if( cameraType == FLY )
		pos += right * units;
}

void Camera::fly(float units)
{
	// move only on y-axis for land object
	if( cameraType == DRIVE )
		pos[1] += units;

	if( cameraType == FLY )
		pos += up * units;
}

void Camera::pitch(float angle)
{
	// rotate around own right vector (x-axis)
	glm::mat4 T;
	float radian = DEGTORAD(angle);
	T = glm::rotate(glm::mat4(1.0f), radian, right);

	// rotate up_ and look_ around right_ vector
	glm::vec4 tUp = T * glm::vec4(up, 0.0f);
	up[0] = tUp[0]; up[1] = tUp[1]; up[2] = tUp[2];
	glm::vec4 tLook = T * glm::vec4(look, 0.0f);
	look[0] = tLook[0]; look[1] = tLook[1]; look[2] = tLook[2];
}

void Camera::yaw(float angle)
{
	glm::mat4 T;
	float radian = DEGTORAD(angle);

	// rotate around world y (0, 1, 0) always for land object (y-axis)
	if( cameraType == DRIVE )
		T = glm::rotate(glm::mat4(1.0f), radian, glm::vec3(0.0f, 1.0f, 0.0f));

	// rotate around own up vector for aircraft
	if( cameraType == FLY )
		T = glm::rotate(glm::mat4(1.0f), radian, up);

	// rotate _right and _look around _up or y-axis
	glm::vec4 tRight = T * glm::vec4(right, 0.0f);
	right[0] = tRight[0]; right[1] = tRight[1]; right[2] = tRight[2];
	glm::vec4 tLook = T * glm::vec4(look, 0.0f);
	look[0] = tLook[0]; look[1] = tLook[1]; look[2] = tLook[2];
}

void Camera::roll(float angle)
{
	float radian = DEGTORAD(angle);
	// only roll for aircraft type (z-axis)
	if( cameraType == FLY )
	{
		glm::mat4 T;
		T = glm::rotate(glm::mat4(1.0f), radian, look);

		// rotate _up and _right around _look vector
		glm::vec4 tRight = T * glm::vec4(right, 0.0f);
		right[0] = tRight[0]; right[1] = tRight[1]; right[2] = tRight[2];
		glm::vec4 tUp = T * glm::vec4(up, 0.0f);
		up[0] = tUp[0]; up[1] = tUp[1]; up[2] = tUp[2];
	}
}

glm::mat4 Camera::lookAt(glm::vec3& eye, glm::vec3& at, glm::vec3& up)
{
	glm::vec3 n, u, v;

	n = eye - at;
	n = glm::normalize(n);

	u = glm::cross(up, n);
	u = glm::normalize(u);

	v = glm::cross(n, u);
	v = glm::normalize(v);

#ifdef DEBUG
	std::cout << "U: " << u[0] << " " << u[1] << " " << u[2] << std::endl;
	std::cout << "V: " << v[0] << " " << v[1] << " " << v[2] << std::endl;
	std::cout << "N: " << n[0] << " " << n[1] << " " << n[2] << std::endl;
#endif

	pos   = eye;
	right = u;
	up    = v;
	look  = n;
	V = View();

	return V;
}

glm::mat4 Camera::View(void) 
{
	// keep camera's axes orthogonal to each other
	look = glm::normalize(look);
	up = glm::normalize(up);

	right = glm::cross(up, look);
	right = glm::normalize(right);
	
	up = glm::cross(look, right);
	up = glm::normalize(up);

	V[0][0] = right[0];	V[1][0] = right[1];	V[2][0] = right[2];	V[3][0] = -glm::dot(right, pos);
	V[0][1] = up[0];	V[1][1] = up[1];	V[2][1] = up[2];	V[3][1] = -glm::dot(up, pos);
	V[0][2] = look[0];	V[1][2] = look[1];	V[2][2] = look[2];	V[3][2] = -glm::dot(look, pos);
	V[0][3] = 0.0;		V[1][3] = 0.0;		V[2][3] = 0.0;		V[3][3] = 1.0f;

#ifdef DEBUG
	printf("\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n\n",
            V[0][0], V[1][0], V[2][0], V[3][0],
	        V[0][1], V[1][1], V[2][1], V[3][1],
	        V[0][2], V[1][2], V[2][2], V[3][2],
	        V[0][3], V[1][3], V[2][3], V[3][3]);
#endif

	return V;
}

} // namespace